           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Domain_SF_prod = Domain_SF + RS_prod:

(*****************************************************************

         1. 
         2. 
         3. 
         4. 

 *****************************************************************)

(***********************************************************
              type def (Product Stable Failure)
 ***********************************************************)

types ('i,'a) dom_SF_prod = "'i => 'a dom_SF"     (* synonym *)

(**************************************************************)
(*                                                            *)
(*  It has already proven that ('i,'a) dom_SF_prod is a CMS   *)
(*  because dom_SF is a CMS and "=>" preserves CMS.           *)
(*                                                            *)
(*    instance dom_SF :: (type) cms_rs  in Domain_SF.thy      *)
(*    instance => :: (type,cms_rs) cms_rs in RS_prod.thy      *)
(*                                                            *)
(*  This can be confirmed as the following lemmas.            *)
(*                                                            *)
(**************************************************************)

lemma limitSF_prod_limit:
  "ALL (SFps::(('i,'a) dom_SF_prod) infinite_seq).
     cauchy SFps --> SFps convergeTo limit SFps"
apply (simp add: the_limit)
      (* see "the_limit" in RS.thy *)
done

(*----------------------------------------------------------*
 |                                                          |
 |         Fixed point induction for refinement             |
 |                                                          |
 *----------------------------------------------------------*)

consts
  Ref_SF_prod :: "('i,'a) dom_SF_prod => ('i,'a) dom_SF_prod => bool"
  Rev_SF_prod :: "('i,'a) dom_SF_prod => ('i,'a) dom_SF_prod => bool"

defs
  Ref_SF_prod_def : "Ref_SF_prod SEp == (%SFp. (SEp <= SFp))"
  Rev_SF_prod_def : "Rev_SF_prod SEp == (%SFp. (SFp <= SEp))"

(************************************************************
             continuity lemma for refinement
 ************************************************************)

lemma Ref_SF_prod_cont: "ALL SEp. continuous (Ref_SF_prod SEp)"
apply (intro allI)
apply (simp add: Ref_SF_prod_def)
apply (simp add: order_prod_def)

apply (insert prod_continuous)
apply (drule_tac x="(%i. Ref_SF (SEp i))" in spec)
apply (simp add: Ref_SF_cont)
apply (simp add: Ref_SF_def)
done

(************************************************************
         continuity lemma for (Reverse) refinement
 ************************************************************)

lemma Rev_SF_prod_cont: "ALL SEp. continuous (Rev_SF_prod SEp)"
apply (intro allI)
apply (simp add: Rev_SF_prod_def)
apply (simp add: order_prod_def)

apply (insert prod_continuous)
apply (drule_tac x="(%i. Rev_SF (SEp i))" in spec)
apply (simp add: Rev_SF_cont)
apply (simp add: Rev_SF_def)
done

(************************************************************
          Metric Fixed Point Induction for refinement
 ************************************************************)

theorem metric_fixpoint_induction_ref:
  "ALL f (SEp::('i,'a) dom_SF_prod) SFp.
     constructive f & mono f & SEp <= f SEp & SFp = f SFp --> SEp <= SFp"
apply (intro allI impI)
apply (insert metric_fixpoint_induction)
apply (drule_tac x="(Ref_SF_prod SEp)" in spec)
apply (drule_tac x="f" in spec)
apply (drule mp)
apply (simp)

      (* satisfiablity *)
apply (rule conjI)
apply (rule_tac x="SEp" in exI)
apply (simp add: Ref_SF_prod_def)

      (* continuity *)
apply (simp add: Ref_SF_prod_cont)

      (* inductiveness *)
apply (simp add: inductivefun_def Ref_SF_prod_def)
apply (intro allI impI)
apply (simp add: mono_def)
apply (erule conjE)+
apply (drule_tac x="SEp" in spec)
apply (drule_tac x="x" in spec)
apply (simp)

      (* fixped point induction *)
apply (elim conjE exE)
apply (simp add: Ref_SF_prod_def)
done

(************************************************************
        Metric Fixed Point Induction for refinement (rev)
 ************************************************************)

theorem metric_fixpoint_induction_rev:
  "ALL f (SEp::('i,'a) dom_SF_prod) SFp.
     constructive f & mono f & f SEp <= SEp & SFp = f SFp --> SFp <= SEp"
apply (intro allI impI)
apply (insert metric_fixpoint_induction)
apply (drule_tac x="(Rev_SF_prod SEp)" in spec)
apply (drule_tac x="f" in spec)
apply (drule mp)
apply (simp)

      (* satisfiablity *)
apply (rule conjI)
apply (rule_tac x="SEp" in exI)
apply (simp add: Rev_SF_prod_def)

      (* continuity *)
apply (simp add: Rev_SF_prod_cont)

      (* inductiveness *)
apply (simp add: inductivefun_def Rev_SF_prod_def)
apply (intro allI impI)
apply (simp add: mono_def)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="SEp" in spec)
apply (simp)

      (* fixped point induction *)
apply (elim conjE exE)
apply (simp add: Rev_SF_prod_def)
done

end
