           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Norm_seq = CMS:

(*****************************************************************

         1. Definition of Normarized sequences
         2. Properties of Normarized sequences
         3. How to transform each Cauchy sequence to NF
         4. The same limit between xs and NF(xs)

 *****************************************************************)

consts
  normal     :: "'a::ms0 infinite_seq => bool" 
  Nmin_check :: "'a::ms0 infinite_seq => real => nat => bool" 
  Nmin       :: "'a::ms0 infinite_seq => real => nat" 
  NF         :: "'a::ms0 infinite_seq => 'a::ms0 infinite_seq"

defs
  normal_def : 
    "normal xs == ALL (n::nat) (m::nat). 
        distance(xs n, xs m) < (1/2)^(min n m)"

  Nmin_check_def :
    "Nmin_check xs delta N == min_check 
     (%N. ALL n m. (N <= m & N <= n) --> distance(xs n, xs m) < delta) N"
  
  Nmin_def :
    "Nmin xs delta == THE N. Nmin_check xs delta N"

  NF_def :
    "NF xs == (%n. xs (Nmin xs ((1/2)^n)))"

(********************************************************************
                          Normalization
 ********************************************************************)

(*** normalized sequence --> Cauchy sequence ***)

lemma normal_cauchy: 
  "ALL xs. normal xs --> cauchy xs"
apply (intro allI impI)
apply (simp add: cauchy_def)
apply (intro allI impI)

apply (insert pow_convergence)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="delta" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (simp add: normal_def)
apply (drule_tac x="i" in spec)
apply (drule_tac x="j" in spec)

apply (insert min_x_y_or)
apply (drule_tac x="i" in spec)
apply (drule_tac x="j" in spec)
apply (erule disjE)

apply (insert ALL_realpow_le_le)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="i" in spec)
apply (simp)

apply (drule_tac x="1/2" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="j" in spec)
apply (simp)
done


lemma normal_limit: 
  "ALL (xs::'a::ms infinite_seq) y. 
     (normal xs & xs convergeTo y) 
         --> (ALL n. distance(xs (Suc n), y) < (1/2)^n)"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (elim conjE)
apply (drule_tac x="(1/2)^(Suc n)" in spec)
apply (simp)
apply (erule exE)
apply (rename_tac xs y n N)

apply (case_tac "N <= Suc n")
 apply (drule_tac x="Suc n" in spec)
 apply (simp add: symmetry_ms)

 apply (insert ALL_real_mult_right_mono)
 apply (drule_tac x="1" in spec)
 apply (drule_tac x="2" in spec)
 apply (drule_tac x="distance (y, xs (Suc n))" in spec)
 apply (drule mp)
 apply (simp add: positive_ms)
 apply (simp)

(* else (i.e. Suc n < N *)
 apply (drule_tac x="N" in spec)
 apply (simp)
 apply (insert triangle_inequality_ms)
 apply (drule_tac x="xs (Suc n)" in spec)
 apply (drule_tac x="xs N" in spec)
 apply (drule_tac x="y" in spec)

 apply (simp add: normal_def)
 apply (drule_tac x="Suc n" in spec)
 apply (drule_tac x="N" in spec)
 apply (simp add: min_x_y)
 apply (simp add: symmetry_ms)
done

(********************************************************************
                                Nmin
 ********************************************************************)

(*** Nmin exists ***)

lemma Nmin_exists: 
  "ALL (xs::'a::ms infinite_seq) (delta::real).
          (0 < delta & cauchy xs) --> (EX N. Nmin_check xs delta N)"
apply (intro allI impI)
apply (erule conjE)
apply (simp add: cauchy_def)
apply (drule_tac x="delta" in spec)
apply (simp)
apply (erule exE)

apply (insert exists_min_nat)
apply (drule_tac x="(%N. ALL n m. N <= m & N <= n 
                         --> distance (xs n, xs m) < delta)" in spec)
apply (rotate_tac -1)
apply (drule_tac x="n" in spec)
apply (drule mp)
apply (simp)

apply (erule exE)
apply (rule_tac x="m" in exI)
apply (simp add: Nmin_check_def)
done

(*** Nmin unique ***)

lemma Nmin_unique: 
  "ALL (xs::'a::ms infinite_seq) (delta::real) N M.
        (Nmin_check xs delta N & Nmin_check xs delta M) --> N = M"
apply (simp add: Nmin_check_def unique_min_nat)
done

(*** the Nmin ***)

lemma the_Nmin:
  "ALL (xs::'a::ms infinite_seq) (delta::real).
     (0 < delta & cauchy xs) --> Nmin_check xs delta (Nmin xs delta)"
apply (intro allI impI)
apply (insert Nmin_exists)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta" in spec)
apply (simp)
apply (erule exE)

apply (insert Nmin_unique)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta" in spec)
apply (drule_tac x="x" in spec)
apply (rename_tac xs delta N)

apply (simp add: Nmin_def)

apply (insert ALL_theI)
apply (drule_tac x="Nmin_check xs delta" in spec)
apply (rotate_tac -1)
apply (drule_tac x="N" in spec)
apply (drule mp)

apply (rule conjI)
apply (drule_tac x="N" in spec)  (* to avoide a loop *)
apply (simp)

apply (rule allI)
apply (drule_tac x="x" in spec)
apply (simp)
apply (simp)
done

(*** the Nmin uniueq ***)

lemma the_Nmin_unique:
  "ALL (xs::'a::ms infinite_seq) (delta::real) n.
     (0 < delta & cauchy xs & Nmin_check xs delta n)
        --> Nmin xs delta = n"
apply (intro allI impI)
apply (elim conjE)
apply (insert the_Nmin)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta" in spec)
apply (simp)
apply (insert Nmin_unique)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="Nmin xs delta" in spec)
apply (simp)
done

(*** the property of Nmin_check execpt "min" ***)

lemma Nmin_check_except_min:
  "ALL (xs::'a::ms infinite_seq) (delta::real) N.
     (0 < delta & cauchy xs & Nmin_check xs delta N) -->
        (ALL n m. (N <= m & N <= n) --> distance(xs n, xs m) < delta)"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: Nmin_check_def)
apply (unfold min_check_def)
apply (elim conjE)
apply (drule_tac x="n" in spec)
apply (rotate_tac -1)
apply (drule_tac x="m" in spec)
apply (simp)
done

(*** the property of Nmin execpt "min" ***)

lemma Nmin_except_min:
  "ALL (xs::'a::ms infinite_seq) (delta::real).
     (0 < delta & cauchy xs)
        --> (ALL n m. (Nmin xs delta <= m & Nmin xs delta <= n) 
           --> distance(xs n, xs m) < delta)"
apply (intro allI impI)
apply (elim conjE)
apply (insert Nmin_check_except_min)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta" in spec)
apply (drule_tac x="Nmin xs delta" in spec)
apply (simp add: the_Nmin)
done


(*****************************
      min_number_cauchy
 *****************************)

(*** Nmin order (check) ***)

lemma min_number_cauchy_check:
  "ALL (xs::'a::ms infinite_seq) (delta1::real) delta2 n1 n2.
     (0 < delta1 & delta1 <= delta2 & cauchy xs &
        Nmin_check xs delta1 n1 & Nmin_check xs delta2 n2)
           --> n2 <= n1"
apply (intro allI impI)
apply (simp add: Nmin_check_def)
apply (unfold min_check_def)
apply (elim conjE)
apply (rotate_tac -1)
apply (drule_tac x="n1" in spec)
apply (rotate_tac -1)
apply (drule mp)

apply (intro allI impI)

apply (drule_tac x="n" in spec)
apply (rotate_tac -1)
apply (drule_tac x="m" in spec)

apply (simp_all)
done

(*** Nmin order ***)

lemma min_number_cauchy:
  "ALL (xs::'a::ms infinite_seq) (delta1::real) delta2.
     (0 < delta1 & delta1 <= delta2 & cauchy xs)
        --> Nmin xs delta2 <= Nmin xs delta1"
apply (intro allI impI)
apply (insert min_number_cauchy_check)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta1" in spec)
apply (drule_tac x="delta2" in spec)
apply (drule_tac x="Nmin xs delta1" in spec)
apply (drule_tac x="Nmin xs delta2" in spec)
apply (insert ALL_order_less_le_trans)
apply (drule_tac x="0" in spec)
apply (drule_tac x="delta1" in spec)
apply (drule_tac x="delta2" in spec)
apply (simp add: the_Nmin)
done


(*** Nmin order half ***)

lemma min_number_cauchy_half:
  "ALL (xs::'a::ms infinite_seq) n m.
     (n <= m & cauchy xs) --> Nmin xs ((1/2)^n) <= Nmin xs ((1/2)^m)"
apply (intro allI impI)
apply (insert ALL_realpow_le_le)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: min_number_cauchy)
done

(*****************************
    normal_form_seq_normal
 *****************************)

lemma normal_form_seq_normal: 
  "ALL (xs::'a::ms infinite_seq). cauchy xs --> normal (NF(xs))"
apply (intro allI impI)
apply (simp add: normal_def NF_def)
apply (intro allI)

apply (case_tac "n <= m")
 apply (simp add: min_x_y)
 apply (simp add: Nmin_except_min min_number_cauchy_half)

(* else *)
 apply (simp add: min_y_x)
 apply (simp add: Nmin_except_min min_number_cauchy_half)
done

(*****************************
   normal_form_seq_same_limit
 *****************************)

(*** only if part ***)

lemma normal_form_seq_same_limit_only_if:
  "ALL (xs::'a::ms infinite_seq) y. 
     (cauchy xs & xs convergeTo y) --> NF(xs) convergeTo y"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (intro allI impI)
apply (elim conjE)
apply (drule_tac x="eps/2" in spec)
apply (simp)
apply (erule exE)

apply (insert pow_convergence)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="eps/2" in spec)
apply (simp)
apply (erule exE)
apply (rename_tac xs y eps N M)

apply (rule_tac x="M" in exI)
apply (intro allI impI)

apply (case_tac "N <= Nmin xs ((1/2)^m)")

 apply (drule_tac x="Nmin xs ((1/2)^m)" in spec)
 apply (simp add: NF_def)

(* else *)
 apply (insert triangle_inequality_ms)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="xs N" in spec)
 apply (drule_tac x="(NF xs) m" in spec)

 apply (drule_tac x="N" in spec)
 apply (simp add: NF_def)

 apply (insert Nmin_except_min)
 apply (drule_tac x="xs" in spec)
 apply (drule_tac x="(1 / 2) ^ m" in spec)
 apply (simp)

 apply (drule_tac x="N" in spec)
 apply (drule_tac x="Nmin xs ((1 / 2) ^ m)" in spec)
 apply (simp)
 apply (insert ALL_realpow_le_le)
 apply (drule_tac x="1/2" in spec)
 apply (drule_tac x="M" in spec)
 apply (drule_tac x="m" in spec)
 apply (simp)
done

(*** if part ***)

lemma normal_form_seq_same_limit_if:
  "ALL (xs::'a::ms infinite_seq) y. 
     (cauchy xs & NF(xs) convergeTo y) --> xs convergeTo y"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (intro allI impI)
apply (erule conjE)
apply (drule_tac x="eps/2" in spec)
apply (simp)
apply (erule exE)

apply (insert pow_convergence)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="eps/2" in spec)
apply (simp)
apply (erule exE)
apply (rename_tac xs y eps N M)

apply (rule_tac x="Nmin xs ((1/2)^(max N M))" in exI)
apply (intro allI impI)

apply (insert triangle_inequality_ms)
apply (drule_tac x="y" in spec)
apply (drule_tac x="xs (Nmin xs ((1/2)^(max N M)))" in spec)
apply (drule_tac x="xs m" in spec)

apply (drule_tac x="max N M" in spec)
apply (simp add: le_maxI1)
apply (simp add: NF_def)

(* *)

apply (insert Nmin_except_min)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="(1/2)^(max N M)" in spec)
apply (simp)

apply (drule_tac x="Nmin xs ((1 / 2) ^ max N M)" in spec)
apply (drule_tac x="m" in spec)
apply (simp)

apply (insert ALL_realpow_le_le)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="M" in spec)
apply (drule_tac x="max N M" in spec)
apply (simp add: le_maxI2)
done

(*** iff ***)

lemma normal_form_seq_same_limit:
  "ALL (xs::'a::ms infinite_seq) y. 
     cauchy xs --> (xs convergeTo y = NF(xs) convergeTo y)"
apply (intro allI impI)
apply (rule iffI)
apply (simp add: normal_form_seq_same_limit_only_if)
apply (insert normal_form_seq_same_limit_if)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
done

end






