           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory RS = Norm_seq:

(*****************************************************************

         1. Definition of Restriction Spaces (RS)
         2. RS ==> MS
         3. Metric Fixed point induction
         4. 

 *****************************************************************)

consts 
  restriction :: "'a::ms0 => nat => 'a::ms0"  ("_ rest _" [55,56] 55)

(********************************************************** 
                  restriction space (rs)
 **********************************************************)

axclass rs < ms0
  zero_eq_rs: "ALL (x::'a::ms0) (y::'a::ms0). x rest 0 = y rest 0"
  min_rs    : "ALL (x::'a::ms0) (m::nat) (n::nat). 
                     (x rest m) rest n = x rest (min m n)"
  diff_rs   : "ALL (x::'a::ms0) (y::'a::ms0).
                     (x ~= y) --> (EX n. x rest n ~= y rest n)"

consts
  distance_rs_check  :: "('a::ms0 * 'a::ms0) => real => bool"
  distance_nat_check :: "('a::ms0 * 'a::ms0) => nat => bool"

defs
  distance_rs_check_def : 
     "distance_rs_check xy r == 
     r |<< {r. EX n. r=(1/2)^n & ((fst xy) rest n) = ((snd xy) rest n)}"
  distance_nat_check_def : 
     "distance_nat_check xy n == 
         max_check (%n. ((fst xy) rest n) = ((snd xy) rest n)) n"

defs (overloaded)
  distance_rs_def : "distance xy == THE r. distance_rs_check xy r"


(********************************************************** 
             complete restriction space (crs)
 **********************************************************)

axclass ms_rs < ms, rs
axclass cms_rs < cms, ms_rs


(********************************************************** 
                        RS lemmas 
 **********************************************************)

(*** contraposition of diff_rs ***)

lemma contra_diff_rs: 
    "ALL (x::'a::rs) (y::'a::rs).
          (ALL n. x rest n = y rest n) --> (x = y)"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (simp add: diff_rs)
done

(*** diff_rs inv ***)

lemma contra_diff_rs_inv: 
    "ALL (x::'a::rs) (y::'a::rs) n.
           x rest n ~= y rest n --> (x ~= y)"
apply (auto)
done

(*** zero distance ***)

lemma distance_zero: 
  "ALL (x::'a::rs). distance_rs_check (x,x) 0"
apply (intro allI)
apply (simp add: distance_rs_check_def)
apply (simp add: pow_glb)
done

(*****************************
        equal-preserv
 *****************************)

lemma rest_equal_preserve:
  "ALL (x::'a::rs) (y::'a::rs) (m::nat) (n::nat).
     (x rest n = y rest n & m <= n) --> x rest m = y rest m"
apply (intro allI impI)
apply (insert min_rs)
apply (drule_tac x="x" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (insert min_rs)
apply (drule_tac x="y" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: min_y_x)
done

lemma rest_nonequal_preserve:
  "ALL (x::'a::rs) (y::'a::rs) (m::nat) (n::nat).
     (x rest m ~= y rest m & m <= n) --> (x rest n ~= y rest n)"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (simp add: disj_not2)
apply (intro impI)
apply (insert rest_equal_preserve)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="n" in spec)
apply (simp)
done

(*****************************
      distance_nat_check
 *****************************)

(*** except max ***)

lemma distance_nat_check_except_max:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     distance_nat_check (x,y) n --> (x rest n = y rest n)"
apply (simp add: distance_nat_check_def max_check_def)
done

(*** only max ***)

lemma distance_nat_check_only_max:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     distance_nat_check (x,y) n --> 
        (ALL m. (x rest m = y rest m) --> m <= n)"
apply (simp add: distance_nat_check_def max_check_def)
done

(*** Suc n => neq ***)

lemma distance_nat_Suc_n:
  "ALL (n::nat) (x::'a::rs) (y::'a::rs).
      distance_nat_check (x, y) n --> x rest Suc n ~= y rest Suc n"
apply (intro allI impI)

apply (case_tac "x rest (Suc n) = y rest (Suc n)")
 apply (insert distance_nat_check_only_max)
 apply (drule_tac x="x" in spec)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="n" in spec)
 apply (simp)
 apply (drule_tac x="Suc n" in spec)
 apply (simp_all)
done

(*****************************
          rest-nat
 *****************************)

(*** max existency ***)

lemma rest_nat_exists:
  "ALL (x::'a::rs) (y::'a::rs).
      x ~= y --> (EX m. distance_nat_check (x,y) m)"
apply (intro allI impI)

apply (insert exists_min_nat)
apply (drule_tac x="(%m. x rest m ~= y rest m)" in spec)
apply (simp add: diff_rs)
apply (erule exE)
apply (rename_tac x y m)
apply (simp add: min_check_def)

apply (insert zero_eq_rs)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)

apply (insert nat_pos_zero)
apply (drule_tac x="m" in spec)
apply (erule disjE)
apply (simp)
apply (erule exE)
apply (rename_tac x y m n)
apply (simp)

 (* n must be the max *)

apply (rule_tac x="n" in exI)
apply (simp add: distance_nat_check_def max_check_def)
apply (rule conjI)

apply (case_tac "x rest n = y rest n")
 apply (simp)

(* otherwise *)

 apply (erule conjE)
 apply (drule_tac x="n" in spec)
 apply (simp)

(* max check *)

apply (intro allI impI)
apply (rename_tac x y m n n')

apply (case_tac "Suc n <= n'")

 apply (insert rest_nonequal_preserve)
 apply (drule_tac x="x" in spec)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="Suc n" in spec)
 apply (drule_tac x="n'" in spec)
 apply (drule mp)
 apply (simp_all)
done

(*** GLB and Max ***)

lemma rest_nat_glb_max:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     distance_nat_check (x,y) n --> distance_rs_check (x,y) ((1 / 2) ^ n)"
apply (intro allI impI)

apply (simp add: distance_rs_check_def is_glb_def is_lb_def)
apply (rule conjI)

(* lb *)

apply (intro allI impI)
apply (erule exE)
apply (simp add: distance_nat_check_def max_check_def)
apply (erule conjE)

apply (drule_tac x="na" in spec)
apply (simp add: ALL_realpow_le_le)

(* glb *)

apply (intro allI impI)
apply (drule_tac x="(1 / 2) ^ n" in spec)
apply (drule mp)
apply (rule_tac x="n" in exI)
apply (simp add: distance_nat_check_def max_check_def)
apply (simp)
done

(*** real <--> nat ***)

lemma rest_nat: 
  "ALL (x::'a::rs) (y::'a::rs).
     x ~= y --> ((EX (m::nat). (distance_rs_check (x,y) ((1 / 2) ^ m))
                                 & distance_nat_check (x,y) m))"
apply (intro allI impI)
apply (insert rest_nat_exists)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="xa" in exI)
apply (simp add: rest_nat_glb_max)
done

(*****************************
      distance existency
 *****************************)

lemma distance_exists: 
  "ALL (x::'a::rs) (y::'a::rs).
        (EX r. 0 <= r & distance_rs_check (x,y) r)"
apply (intro allI)

apply (case_tac "x = y")
 apply (rule_tac x="0" in exI)
 apply (simp add: distance_zero)

(* otherwise (i.e. x ~= y) *)

 apply (insert rest_nat)
 apply (drule_tac x="x" in spec)
 apply (drule_tac x="y" in spec)
 apply (simp)
 apply (erule exE)
 apply (rule_tac x="(1 / 2) ^ m" in exI)
 apply (simp add: zero_le_power)
done

(*****************************
        distance unique
 *****************************)

lemma distance_unique: 
  "ALL (x::'a::rs) (y::'a::rs) (r::real) (s::real).
       (distance_rs_check (x,y) r & distance_rs_check (x,y) s) --> r=s"
apply (intro allI impI)
apply (simp add: distance_rs_check_def)
apply (simp add: glb_unique)
done

(*****************************
        the distance
 *****************************)

lemma the_distance:
  "ALL (x::'a::rs) (y::'a::rs).
     distance_rs_check (x,y) (distance (x,y))"
apply (intro allI)
apply (insert distance_exists)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (erule exE)

apply (insert distance_unique)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="r" in spec)
apply (simp)

apply (simp add: distance_rs_def)

apply (insert ALL_theI)
apply (drule_tac x="distance_rs_check (x, y)" in spec)
apply (rotate_tac -1)
apply (drule_tac x="r" in spec)
apply (simp)
done

(*** the Nmin uniueq ***)

lemma the_distance_unique:
  "ALL (x::'a::rs) (y::'a::rs) (r::real).
     distance_rs_check (x,y) r --> distance (x,y) = r"
apply (intro allI impI)
apply (insert the_distance)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (insert distance_unique)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="distance (x, y)" in spec)
apply (drule_tac x="r" in spec)
apply (simp)
done

(******************************************
               THE rest_nat
 ******************************************)

lemma the_rest_nat: 
  "ALL (x::'a::rs) (y::'a::rs).
     x ~= y --> (EX (m::nat). (distance(x,y) = ((1 / 2) ^ m)
                             & distance_nat_check (x,y) m))"
apply (intro allI impI)
apply (insert rest_nat)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="m" in exI)
apply (simp add: the_distance_unique)
done

(********************************************************** 
                       rs --> ms
 **********************************************************)

(*** positive_rs ***)

lemma positive_rs:
  "ALL (x::'a::rs) (y::'a::rs). 0 <= distance (x,y)"
apply (intro allI)
apply (insert distance_exists)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (erule exE)
apply (auto simp add: the_distance_unique)
done

(*** diagonal_rs ***)

lemma diagonal_rs_lm:
  "ALL (x::'a::rs) (y::'a::rs). distance_rs_check (x, y) 0 --> x=y"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (insert rest_nat)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (erule exE)
apply (insert distance_unique)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="(1 / 2) ^ m" in spec)
apply (drule_tac x="0" in spec)
apply (auto)
done

lemma diagonal_rs:
    "ALL (x::'a::rs) (y::'a::rs). (distance(x,y) = 0) = (x = y)"
apply (intro allI)
apply (insert the_distance)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)

apply (rule iffI)
apply (simp add: diagonal_rs_lm)

apply (insert distance_zero)
apply (drule_tac x="y" in spec)
apply (simp add: the_distance_unique)
done

lemma same_pnt_zero_rs:
    "ALL (x::'a::rs) (y::'a::rs). distance(x,x) = 0"
apply (simp add: diagonal_rs)
done

(*** symmetry_rs ***)

lemma symmetry_rs_check_lm:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
        (x rest n = y rest n) = (y rest n = x rest n)"
apply (auto)
done

lemma symmetry_rs_check:
  "ALL (x::'a::rs) (y::'a::rs) (r::real).
        distance_rs_check (x,y) r = distance_rs_check (y,x) r"
apply (intro allI)
apply (simp add: distance_rs_check_def)
apply (simp add: symmetry_rs_check_lm)
done

lemma symmetry_rs:
  "ALL (x::'a::rs) (y::'a::rs). distance(x,y) = distance(y,x)"
apply (intro allI)
apply (insert the_distance)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (insert the_distance)
apply (drule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (insert distance_unique)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="distance (x, y)" in spec)
apply (drule_tac x="distance (y, x)" in spec)
apply (simp add: symmetry_rs_check)
done

(*** triangle_inequality_rs ***)

lemma triangle_inequality_rs_max_ne:
  "ALL (x::'a::rs) (y::'a::rs) (z::'a::rs).
       (x ~= y & y ~= z & x ~= z)
           --> distance (x,z) <= max (distance (x,y)) (distance (y,z))"
apply (intro allI impI)
apply (elim conjE)

apply (insert the_rest_nat)
apply (frule_tac x="x" in spec)
apply (frule_tac x="y" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="z" in spec)
apply (drule_tac x="z" in spec)
apply (drule_tac x="y" in spec)
apply (simp)

apply (elim exE conjE)
apply (rename_tac x y z k n m)
apply (simp)

(* FOR x rest min m n = z rest min m n *)

apply (insert min_rs)
apply (frule_tac x="x" in spec)
apply (frule_tac x="y" in spec)
apply (frule_tac x="y" in spec)
apply (drule_tac x="z" in spec)

apply (drule_tac x="m" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="n" in spec)

apply (drule_tac x="n" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="m" in spec)

apply (simp add: min_commute)

(* FOR distance --> (1/2)^n *)

apply (insert distance_nat_check_except_max)
apply (frule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="z" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="n" in spec)
apply (simp)

apply (insert distance_nat_check_only_max)
apply (drule_tac x="x" in spec)
apply (drule_tac x="z" in spec)
apply (drule_tac x="k" in spec)
apply (simp)
apply (drule_tac x="min m n" in spec)
apply (simp)

apply (insert ALL_realpow_le_le)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="min m n" in spec)
apply (drule_tac x="k" in spec)
apply (simp add: pow_min_max_dist)
done

lemma triangle_inequality_rs_max:
  "ALL (x::'a::rs) (y::'a::rs) (z::'a::rs).
         distance (x,z) <= max (distance (x,y)) (distance (y,z))"
apply (intro allI)

apply (case_tac "x = y")
 apply (simp add: same_pnt_zero_rs positive_rs max_def)

 apply (case_tac "y = z")
  apply (simp add: same_pnt_zero_rs positive_rs max_def)

  apply (case_tac "x = z")
  apply (simp add: same_pnt_zero_rs positive_rs max_def)

apply (simp add: triangle_inequality_rs_max_ne)
done

lemma triangle_inequality_rs:
  "ALL (x::'a::rs) (y::'a::rs) (z::'a::rs).
         distance (x,z) <= distance (x,y) + distance (y,z)"
apply (intro allI)
apply (insert triangle_inequality_rs_max)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="z" in spec)
apply (insert real_max_plus)
apply (drule_tac x="distance (x, y)" in spec)
apply (drule_tac x="distance (y, z)" in spec)
apply (drule mp)
apply (simp add: positive_rs)

apply (blast intro: order_trans)
done



(**************************************************************

       (restriction space ==> metric space)

  instance x :: (type) rs ==> instance x :: (type) ms

                 by  positive_rs
                     diagonal_rs
                     symmetry_rs
                     triangle_inequality_rs

 **************************************************************)

(*** ex: ms properties ***)

(*** see diff_pnt_pos ***)

lemma diff_pnt_pos_rs:
  "ALL (x::'a::rs) (y::'a::rs). x ~= y --> 0 < distance(x,y)"
apply (intro allI impI)

apply (case_tac "0 < distance (x, y)")
apply (simp)

(* otherwise --> contradict *)
apply (simp add: positive_rs real_less_le diagonal_rs)
done

(*** see diff_pnt_pos_inv ***)

lemma diff_pnt_pos_rs_inv:
  "ALL (x::'a::rs) (y::'a::rs). 0 < distance(x,y) --> x ~= y"
apply (intro allI impI)

apply (case_tac "x = y")
apply (insert diagonal_rs)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp_all)
done

(*****************************
        rest_nat_le
 *****************************)

(* 1 only if *)

lemma rest_nat_le_1_only_if:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     x rest n = y rest n --> distance(x,y) <= (1/2)^n"
apply (intro allI impI)

apply (case_tac "x = y")
apply (simp add: same_pnt_zero_rs positive_rs zero_le_power)

apply (insert the_rest_nat)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (erule exE)

apply (insert distance_nat_check_only_max)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="m" in spec)
apply (simp)
apply (drule_tac x="n" in spec)
apply (simp add: ALL_realpow_le_le)
done

(* 1 if *)

lemma rest_nat_le_1_if:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
      distance(x,y) <= (1/2)^n --> x rest n = y rest n"
apply (intro allI impI)
apply (case_tac "x = y")
apply (simp add: same_pnt_zero_rs positive_rs zero_le_power)

apply (insert the_rest_nat)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (erule exE)

apply (insert distance_nat_check_except_max)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="m" in spec)
apply (simp)

apply (insert ALL_realpow_le_le_inv)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp)

apply (insert rest_equal_preserve)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp)
done

(* 1 iff *)

lemma rest_nat_le_1:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     (x rest n = y rest n) = (distance(x,y) <= (1/2)^n)"
apply (intro allI)
apply (intro iffI)
apply (simp add: rest_nat_le_1_only_if)
apply (simp add: rest_nat_le_1_if)
done

(* 2 iff *)

lemma rest_nat_le_2:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     (x rest n ~= y rest n) = ((1/2)^n < distance(x,y))"
apply (auto simp add: rest_nat_le_1)
done

(*****************************
       rest_nat_less
 *****************************)

(* 1 only if *)

lemma rest_nat_less_1_only_if:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     x rest (Suc n) = y rest (Suc n) --> distance(x,y) < (1/2)^n"
apply (intro allI intro impI)
apply (simp add: rest_nat_le_1)
apply (simp add: ALL_realpow_le_Suc_less)
done

(* 1 if *)

lemma rest_nat_less_1_if:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     distance(x,y) < (1/2)^n --> x rest (Suc n) = y rest (Suc n)"
apply (intro allI impI)

apply (case_tac "x = y")
apply (simp)

apply (insert the_rest_nat)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (erule exE)
apply (simp)

apply (insert ALL_realpow_less_less_inv)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp)

apply (case_tac "Suc n <= m")
 apply (simp)

 apply (insert distance_nat_check_except_max)
 apply (drule_tac x="x" in spec)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="m" in spec)
 apply (simp)

 apply (insert rest_equal_preserve)
 apply (drule_tac x="x" in spec)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="Suc n" in spec)
 apply (drule_tac x="m" in spec)
 apply (simp)

(* otherwise --> contradiction *)
apply (simp)
done

(* 1 iff *)

lemma rest_nat_less_1:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     (x rest (Suc n) = y rest (Suc n)) = (distance(x,y) < (1/2)^n)"
apply (intro allI iffI)
apply (simp add: rest_nat_less_1_only_if)
apply (simp add: rest_nat_less_1_if)
done

(* 2 iff *)

lemma rest_nat_less_2:
  "ALL (x::'a::rs) (y::'a::rs) (n::nat).
     (x rest (Suc n) ~= y rest (Suc n)) = ((1/2)^n <= distance(x,y))"
apply (auto simp add: rest_nat_less_1)
done

(*****************************
    mini_number_cauchy_rest
 *****************************)

lemma mini_number_cauchy_rest:
  "ALL (xs::'a::rs infinite_seq).
      normal xs --> (ALL n m. k <= n & k <= m 
         --> (xs n) rest k = (xs m) rest k)"
apply (intro allI impI)
apply (simp add: normal_def)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)

apply (insert ALL_realpow_le_le)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="k" in spec)
apply (drule_tac x="min n m" in spec)
apply (simp add: rest_nat_le_1)
done

(*****************************
         rest_limit
 *****************************)

lemma rest_limit:
  "ALL (xs::'a::rs infinite_seq) y.
      (ALL n. y rest n = (xs n) rest n)
           --> xs convergeTo y"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (intro allI impI)

apply (insert pow_convergence)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="eps" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (rename_tac xs y eps N n)

apply (drule_tac x="n" in spec)
apply (simp add: rest_nat_le_1)

apply (insert ALL_realpow_le_le)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="N" in spec)
apply (drule_tac x="n" in spec)

apply (simp)
done

(*****************************
       rest_to_dis_plus
 *****************************)

(*** x1 = x2 --> y1 = y2 ? ***)

lemma contra_diff_rs_plus_N: 
    "ALL (x::'a::rs) (y::'a::rs) (N::nat).
          (ALL n. x rest (n+N) = y rest (n+N)) --> (x = y)"
apply (intro allI impI)
apply (insert contra_diff_rs)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule mp)

apply (intro allI)
apply (drule_tac x="n" in spec)
apply (insert rest_equal_preserve)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="n+N" in spec)
apply (simp_all)
done

(*** x1 ~= x2 --> y1 ~= y2 ***)

lemma rest_to_dis_plus_x_y_neq:
  "ALL (N::nat) (x1::'a::rs) (x2::'a::rs) (y1::'b::rs) (y2::'b::rs).
      (ALL n. ((x1 rest n = x2 rest n) = (y1 rest (n+N) = y2 rest (n+N))))
      & x1 ~= x2 --> y1 ~= y2"
apply (intro allI impI)
apply (insert diff_rs)
apply (drule_tac x="x1" in spec)
apply (drule_tac x="x2" in spec)
apply (auto)
done

(*** 2 maxes --> m = n+N ***)

lemma rest_to_dis_plus_max:
  "ALL (N::nat) (x1::'a::rs) (x2::'a::rs) (y1::'b::rs) (y2::'b::rs) n m.
      (ALL n. ((x1 rest n = x2 rest n) = (y1 rest (n+N) = y2 rest (n+N))))
       & distance_nat_check (x1, x2) n & distance_nat_check (y1, y2) m
          --> m = n+N"
apply (intro allI impI)

apply (rule order_antisym)
apply (insert distance_nat_check_only_max)
apply (drule_tac x="x1" in spec)
apply (drule_tac x="x2" in spec)
apply (rotate_tac -1)
apply (drule_tac x="n" in spec)
apply (simp)

apply (case_tac "m <= N")
 apply (simp)

 apply (drule_tac x="m-N" in spec)
 apply (simp add: distance_nat_check_except_max)
 apply (simp add: ALL_nat_le_diff_plus)

(* otherwise "N < m" *)

 apply (drule_tac x="y1" in spec)
 apply (drule_tac x="y2" in spec)
 apply (drule_tac x="m" in spec)
 apply (simp)
 apply (drule_tac x="n+N" in spec)

 apply (insert distance_nat_check_except_max)
 apply (drule_tac x="x1" in spec)
 apply (drule_tac x="x2" in spec)
 apply (drule_tac x="n" in spec)
 apply (simp)
done

(*** lemma x1 = x2 ***)

lemma rest_to_dis_plus_eq:
  "ALL (N::nat) (x::'a::rs) (y1::'b::rs) (y2::'b::rs).
      (ALL n. (y1 rest (n+N) = y2 rest (n+N)))
          --> (1/2)^N * distance(x,x) = distance(y1,y2)"
apply (intro allI impI)
apply (insert contra_diff_rs_plus_N)
apply (drule_tac x="y1" in spec)
apply (drule_tac x="y2" in spec)
apply (rotate_tac -1)
apply (drule_tac x="N" in spec)
apply (simp add: same_pnt_zero_rs)
done

(*** rest_to_dis_plus ***)

lemma rest_to_dis_plus:
  "ALL (N::nat) (x1::'a::rs) (x2::'a::rs) (y1::'b::rs) (y2::'b::rs).
      (ALL n. ((x1 rest n = x2 rest n) = (y1 rest (n+N) = y2 rest (n+N))))
          --> (1/2)^N * distance(x1,x2) = distance(y1,y2)"
apply (intro allI impI)

apply (case_tac "x1 = x2")
apply (simp add: rest_to_dis_plus_eq)

apply (insert the_rest_nat)
apply (drule_tac x="x1" in spec)
apply (drule_tac x="x2" in spec)
apply (insert the_rest_nat)
apply (drule_tac x="y1" in spec)
apply (drule_tac x="y2" in spec)
apply (simp)
apply (insert rest_to_dis_plus_x_y_neq)
apply (rotate_tac -1)
apply (drule_tac x="N" in spec)
apply (drule_tac x="x1" in spec)
apply (drule_tac x="x2" in spec)
apply (drule_tac x="y1" in spec)
apply (drule_tac x="y2" in spec)
apply (rotate_tac -1)
apply (drule mp)
apply (simp_all)

apply (erule exE)+
apply (rename_tac N x1 x2 y1 y2 n m)
apply (simp)

apply (insert rest_to_dis_plus_max)
apply (rotate_tac -1)
apply (drule_tac x="N" in spec)
apply (drule_tac x="x1" in spec)
apply (drule_tac x="x2" in spec)
apply (drule_tac x="y1" in spec)
apply (drule_tac x="y2" in spec)
apply (rotate_tac -1)
apply (drule_tac x="n" in spec)
apply (rotate_tac -1)
apply (drule_tac x="m" in spec)
apply (simp add: ALL_pow_add)
done

(*************************************************
        rest_to_dis_exists (fun and pair)
 *************************************************)

lemma rest_to_dis_exists_pair:
  "ALL (f::('a => 'c::rs))
       (xps::('a * 'a) set) (y1::'b::rs) (y2::'b::rs).
       xps ~= {} &
      (ALL n. (ALL x:xps. f (fst x) rest n = f (snd x) rest n)
               --> (y1 rest n = y2 rest n))
         --> (EX x. x:xps & distance(y1,y2) <= distance(f (fst x), f (snd x)))"
apply (intro allI impI)
apply (case_tac "y1 = y2")
 apply (simp add: same_pnt_zero_rs positive_rs)
 apply (erule nonempty_exE)
 apply (erule exE)
 apply (rule_tac x="fst x" in exI)
 apply (rule_tac x="snd x" in exI)
 apply (simp add: positive_rs)

       (* y1 ~= y2 *)

 apply (insert the_rest_nat)
 apply (drule_tac x="y1" in spec)
 apply (drule_tac x="y2" in spec)
 apply (simp)
 apply (erule exE)
 apply (simp)

 apply (erule conjE)+
 apply (drule_tac x="Suc m" in spec)
 apply (simp add: distance_nat_Suc_n Bex_def)
 apply (elim conjE exE)
 apply (rule_tac x="a" in exI)
 apply (rule_tac x="b" in exI)
 apply (simp add: rest_nat_less_2 ALL_realpow_le_le)
done

(************************************************************
               distance lemma (different types)
 ************************************************************)

lemma distance_diff_type:
  "ALL (x::'a::rs) (y::'a::rs) (X::'b::rs) (Y::'b::rs).
      (ALL n. (x rest n = y rest n) = (X rest n = Y rest n))
          --> distance(x, y) = distance(X, Y)"
apply (intro allI impI)

apply (case_tac "x = y")
apply (simp)
apply (insert contra_diff_rs)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (simp add: same_pnt_zero_rs)

 apply (drule_tac x="X" in spec)    (* dummy *)
 apply (drule_tac x="X" in spec)    (* dummy *)
 apply (simp)                       (* dummy *)

apply (insert diff_rs)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (erule exE)
apply (frule_tac x="n" in spec)
apply (simp)

apply (insert the_rest_nat)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (insert the_rest_nat)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (insert contra_diff_rs_inv)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (rotate_tac -1)
apply (drule_tac x="n" in spec)
apply (simp)

apply (erule exE)+
apply (rename_tac x y X Y n M N)
apply (simp add: distance_nat_check_def)

 apply (drule_tac x="n" in spec) (* dummy *)
apply (insert unique_max_nat)
apply (drule_tac x="(%n. X rest n = Y rest n)" in spec)
apply (drule_tac x="M" in spec)
apply (drule_tac x="N" in spec)
apply (simp)
done


(*----------------------------------------------------------*
 |                                                          |
 |              Metric Fixed point induction                |
 |                                                          |
 *----------------------------------------------------------*)

consts
  constructive :: "('a::ms0 => 'b::ms0) => bool"
  continuous   :: "('a::ms0 => bool) => bool"
  inductivefun :: "('a::ms0 => bool) => ('a::ms0 => 'a::ms0) => bool"

defs
  constructive_def :
     "constructive f == 
          (ALL x y n. x rest n = y rest n 
               --> (f x) rest (Suc n) = (f y) rest (Suc n))"
  continuous_def :
     "continuous R ==
          (ALL x. ~ R x --> (EX n. ALL y. y rest n = x rest n --> ~ R y))"
 inductivefun_def :
     "inductivefun R f == (ALL x. R x --> R (f x))"


(********************************************************** 
              Construction --> Contraction
 **********************************************************)

lemma contst_to_contra_half:
  "ALL (f::('a::rs => 'b::rs)).
       constructive f --> contraction_alpha f (1/2)"
apply (intro allI impI)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (intro allI)

apply (case_tac "x=y")
apply (simp add: same_pnt_zero_rs)

apply (case_tac "f x = f y")
apply (simp add: same_pnt_zero_rs positive_rs)

apply (insert the_rest_nat)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (insert the_rest_nat)
apply (drule_tac x="f x" in spec)
apply (drule_tac x="f y" in spec)
apply (simp)
apply (elim conjE exE)
apply (simp)

apply (insert distance_nat_check_except_max)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="m" in spec)

apply (simp add: constructive_def)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="m" in spec)
apply (simp)

apply (insert distance_nat_check_only_max)
apply (drule_tac x="f x" in spec)
apply (drule_tac x="f y" in spec)
apply (drule_tac x="ma" in spec)
apply (simp)
apply (drule_tac x="Suc m" in spec)
apply (simp)

apply (insert ALL_realpow_le_le)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="Suc m" in spec)
apply (drule_tac x="ma" in spec)
apply (simp)
done

(*** contractive ***)

lemma contst_to_contra:
  "ALL (f::('a::rs => 'b::rs)).
       constructive f --> contraction f"
apply (intro allI impI)
apply (simp add: contraction_def)
apply (rule_tac x="1/2" in exI)
apply (simp add: contst_to_contra_half)
done

(********************************************************** 
              Construction --> Contraction
 **********************************************************)

lemma contra_half_to_contst:
  "ALL (f::('a::rs => 'b::rs)).
       contraction_alpha f (1/2) --> constructive f"
apply (intro allI impI)

apply (simp add: constructive_def)
apply (intro allI impI)

apply (simp add: contraction_alpha_def map_alpha_def)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)

apply (simp add: rest_nat_le_1)
done

(********************************************************** 
              inducrive function --> ALL n
 **********************************************************)

lemma inductivefun_all_n:
  "ALL R f x. (inductivefun R f & R x)--> (ALL n. R ((f^n) x))"
apply (intro allI impI)
apply (induct_tac n)
apply (simp)
apply (simp add: inductivefun_def)
done

(********************************************************** 
              Metric Fixed Point Induction
 **********************************************************)

theorem metric_fixpoint_induction:
  "ALL (R::('a::cms_rs => bool)) f. 
      (EX x. R x) & continuous R & constructive f & inductivefun R f 
        --> (EX y. (R y 
                    & y = f y
                    & (ALL z. z = f z --> y = z)))"
apply (intro allI impI)
apply (elim conjE exE)
apply (insert Banach)
apply (drule_tac x="f" in spec)
apply (drule_tac x="(%n. ((f^n) x))" in spec)
apply (simp add: contst_to_contra)
apply (elim conjE exE)
apply (rule_tac x="y" in exI)
apply (rule conjI)

apply (drule_tac x="y" in spec)   (* dummy *)
apply (simp add: continuous_def)
apply (drule_tac x="y" in spec)

      (* to use a contradiction *)
apply (case_tac "R y")
apply (simp)
apply (simp)
apply (elim conjE exE)

apply (simp add: convergeTo_def)
apply (drule_tac x="(1/2)^n" in spec)
apply (simp)
apply (elim conjE exE)
apply (drule_tac x="na" in spec)
apply (simp)

apply (insert rest_nat_le_1_if)
apply (rotate_tac -1)
apply (drule_tac x="y" in spec)
apply (rotate_tac -1)
apply (drule_tac x="(f ^ na) x" in spec)
apply (drule_tac x="n" in spec)
apply (simp)

apply (drule_tac x="(f ^ na) x" in spec)
apply (simp add: inductivefun_all_n)
apply (blast)
done


end
