           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory RS_prod = RS:

(*****************************************************************

         1. Productions of RS are also RS.
         2. 
         3. 
         4. 

 *****************************************************************)

(********************************************************** 
              def: prod of restriction space
 **********************************************************)

instance fun :: (type,rs) ms0
apply (intro_classes)
done

defs (overloaded)
  prod_restriction_def : 
    "(xp::('i => 'a::rs)) rest n == (%i. ((xp i) rest n))"

consts
  prod_limit :: "(('i => 'a::ms0) infinite_seq) => ('i => 'a::ms0)"

defs
  prod_limit_def :"prod_limit xps == (%i. limit xps<<i>>)"

(************************************************************
                           Basics
 ************************************************************)

lemma rest_to_prod_rest:
    "ALL (xp::('i => 'a::rs)) (yp::('i => 'a::rs)) n.
         (ALL i. (xp i) rest n = (yp i) rest n) --> xp rest n = yp rest n"
apply (simp add: prod_restriction_def)
done

(************************************************************
                     Product RS ==> RS 
 ************************************************************)

(*** prod_zero_eq_rs ***)

lemma prod_zero_eq_rs:
  "ALL (xp::('i => 'a::rs)) (yp::('i => 'a::rs)).
     xp rest 0 = yp rest 0"
apply (intro allI)
apply (simp add: prod_restriction_def)
apply (simp add: expand_fun_eq)
apply (simp add: zero_eq_rs)
done

(*** prod_min_rs ***)

lemma prod_min_rs: 
  "ALL (xp::('i => 'a::rs)) (m::nat) (n::nat). 
           (xp rest m) rest n = xp rest (min m n)"
apply (intro allI)
apply (simp add: prod_restriction_def)
apply (simp add: expand_fun_eq)
apply (simp add: min_rs)
done

(*** prod_diff_rs ***)

lemma prod_diff_rs: 
  "ALL (xp::('i => 'a::rs)) (yp::('i => 'a::rs)).
                (xp ~= yp) --> (EX n. xp rest n ~= yp rest n)"
apply (intro allI impI)
apply (simp add: prod_restriction_def)

apply (simp add: expand_fun_eq)
apply (erule exE)

apply (insert diff_rs)
apply (drule_tac x="xp x" in spec)
apply (drule_tac x="yp x" in spec)
apply (auto)
done

(*****************************
       Prod RS => RS
 *****************************)

instance fun :: (type,rs) rs
apply (intro_classes)
apply (simp add: prod_zero_eq_rs)
apply (simp add: prod_min_rs)
apply (simp add: prod_diff_rs)
done

(************************************************************
                   Product RS ==> MS 
 ************************************************************)

instance fun :: (type,rs) ms
apply (intro_classes)
apply (simp add: positive_rs)
apply (simp add: diagonal_rs)
apply (simp add: symmetry_rs)
apply (simp add: triangle_inequality_rs)
done

(************************************************************
             i.e.  Product RS ==> MS & RS 
 ************************************************************)

instance fun :: (type,rs) ms_rs
apply (intro_classes)
done

(************************************************************
                     Lemmas (distance)
 ************************************************************)

(*****************************
      prod_distance_sup
 *****************************)

(*** prod_distance (Upper Bound) ***)

lemma prod_distance_ub: 
  "ALL (xp::('i => 'a::rs)) (yp::('i => 'a::rs)).
      {r. EX i. r=distance(xp i, yp i)} <| distance(xp, yp)"
apply (intro allI)
apply (simp add: is_ub_def)
apply (intro allI impI)
apply (erule exE)

apply (case_tac "xp i = yp i")
apply (simp add: same_pnt_zero_rs positive_rs)

apply (insert the_rest_nat)
apply (drule_tac x="xp i" in spec)
apply (drule_tac x="yp i" in spec)
apply (insert the_rest_nat)
apply (drule_tac x="xp" in spec)
apply (drule_tac x="yp" in spec)
apply (simp add: fun_neq)
apply (elim exE)
apply (rename_tac xp yp y i n m)

apply (insert distance_nat_check_except_max)
apply (drule_tac x="xp" in spec)
apply (drule_tac x="yp" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: prod_restriction_def)
apply (simp add: expand_fun_eq)
apply (drule_tac x="i" in spec)

apply (insert distance_nat_check_only_max)
apply (drule_tac x="xp i" in spec)
apply (drule_tac x="yp i" in spec)
apply (drule_tac x="n" in spec)
apply (simp)
apply (drule_tac x="m" in spec)

apply (simp add: ALL_realpow_le_le)
done

(*** prod_distance (Least Upper Bound) ***)

lemma prod_distance: 
  "ALL (xp::('i => 'a::rs)) (yp::('i => 'a::rs)).
      {r. EX i. r=distance(xp i, yp i)} <<| distance(xp, yp)"
apply (intro allI)
apply (simp add: is_lub_def)
apply (simp add: prod_distance_ub)
apply (intro allI impI)

apply (case_tac "xp = yp")
apply (simp add: is_ub_def same_pnt_zero_rs positive_rs)

(* least neq *)

apply (insert the_rest_nat)
apply (drule_tac x="xp" in spec)
apply (drule_tac x="yp" in spec)
apply (simp)
apply (erule exE)
apply (simp)

(* to use a contradiction *)

apply (case_tac "(1 / 2) ^ m <= y")
apply (simp)

      (* --> contradiction *)

 apply (simp add: linorder_not_le)

 apply (insert rest_to_prod_rest)
 apply (drule_tac x="xp" in spec)
 apply (drule_tac x="yp" in spec)
 apply (drule_tac x="Suc m" in spec)
 apply (drule mp)

 apply (intro allI)
 apply (simp add: is_ub_def)
 apply (drule_tac x="distance (xp i, yp i)" in spec)
 apply (drule mp)
 apply (rule_tac x="i" in exI)
 apply (simp)

apply (simp add: rest_nat_less_1)
apply (simp add: rest_nat_less_1)
done

(*****************************
         product maps
 *****************************)

(*** non_expanding ***)

lemma prod_non_expand:
  "ALL (fp::('a::rs => ('i => 'b::rs))).
     (ALL i. non_expanding fp<<i>>) --> non_expanding fp"
apply (intro allI impI)
apply (simp add: non_expanding_def map_alpha_def)
apply (intro allI)
apply (insert prod_distance)
apply (drule_tac x="fp x" in spec)
apply (drule_tac x="fp y" in spec)

apply (simp add: is_lub_def)
apply (erule conjE)
apply (drule_tac x="distance (x, y)" in spec)
apply (drule mp)

apply (simp add: is_ub_def)
apply (intro allI impI)

apply (erule exE)
apply (drule_tac x="i" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp add: exchange_args_def)

apply (simp)
done


(*** alpha_contraction ***)

lemma prod_contraction_alpha:
  "ALL (fp::('a::rs => ('i => 'b::rs))) alpha.
     (ALL i. contraction_alpha fp<<i>> alpha)
         --> contraction_alpha fp alpha"
apply (intro allI impI)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (intro allI)
apply (insert prod_distance)
apply (drule_tac x="fp x" in spec)
apply (drule_tac x="fp y" in spec)

apply (simp add: is_lub_def)
apply (erule conjE)+
apply (drule_tac x="alpha * distance (x, y)" in spec)
apply (drule mp)

apply (simp add: is_ub_def)
apply (intro allI impI)

apply (erule exE)
apply (drule_tac x="i" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)
apply (simp add: exchange_args_def)

apply (simp)
done

(************************************************************
                       Lemmas (limit)
 ************************************************************)

(*** cauchy ***)

lemma prod_cauchy_seq:
  "ALL (xps::(('i => 'b::rs) infinite_seq)).
     cauchy xps --> (ALL i. cauchy xps<<i>>)"
apply (intro allI impI)
apply (simp add: cauchy_def)
apply (intro allI impI)

apply (drule_tac x="delta" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (rename_tac xps i delta N n m)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: exchange_args_def)

apply (insert prod_distance)
apply (drule_tac x="xps n" in spec)
apply (drule_tac x="xps m" in spec)

apply (simp add: is_lub_def)
apply (simp add: is_ub_def)
apply (erule conjE)+
apply (drule_tac x="distance ((xps n) i, (xps m) i)" in spec)
apply (drule mp)

apply (rule_tac x="i" in exI)
apply (simp_all)
done

(*** normal ***)

lemma prod_normal_seq:
  "ALL (xps::(('i => 'b::rs) infinite_seq)).
     normal xps --> (ALL i. normal xps<<i>>)"
apply (intro allI impI)
apply (simp add: normal_def)
apply (intro allI)

apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp add: exchange_args_def)

apply (insert prod_distance)
apply (drule_tac x="xps n" in spec)
apply (drule_tac x="xps m" in spec)

apply (simp add: is_lub_def)
apply (simp add: is_ub_def)
apply (erule conjE)+
apply (drule_tac x="distance ((xps n) i, (xps m) i)" in spec)
apply (drule mp)

apply (rule_tac x="i" in exI)
apply (simp_all)
done

(*** limit (single <-- prod) ***)

lemma prod_limit_from:
  "ALL (xps::(('i => 'b::rs) infinite_seq)) yp.
     xps convergeTo yp --> (ALL i. xps<<i>> convergeTo (yp i))"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (intro allI impI)

apply (drule_tac x="eps" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (rename_tac xps yp i eps N n)
apply (drule_tac x="n" in spec)
apply (simp add: exchange_args_def)

apply (insert prod_distance)
apply (drule_tac x="yp" in spec)
apply (drule_tac x="xps n" in spec)

apply (simp add: is_lub_def)
apply (simp add: is_ub_def)

apply (erule conjE)+
apply (drule_tac x="distance (yp i, (xps n) i)" in spec)
apply (drule mp)

apply (rule_tac x="i" in exI)
apply (simp_all)
done

(*** limit (single --> prod) (note: ms_rs) ***)

lemma prod_limit_to:
  "ALL (xps::(('i => 'b::ms_rs) infinite_seq)) yp.
      (normal xps & (ALL i. xps<<i>> convergeTo (yp i))) 
          --> xps convergeTo yp"
apply (intro allI impI)
apply (simp (no_asm) add: convergeTo_def)
apply (intro allI impI)

apply (insert pow_convergence)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="eps" in spec)
apply (simp)
apply (erule exE)
apply (rename_tac xps yp eps M)

apply (rule_tac x="Suc M" in exI)
apply (intro allI impI)

apply (insert prod_distance)
apply (drule_tac x="yp" in spec)
apply (drule_tac x="xps m" in spec)

apply (simp add: is_lub_def)
apply (erule conjE)+
apply (drule_tac x="(1 / 2) ^ M" in spec)
apply (drule mp)

apply (simp add: is_ub_def)
apply (intro allI impI)
apply (erule exE)

apply (simp)
apply (drule_tac x="i" in spec)

apply (insert Suc_le_nat_exists)
apply (drule_tac x="M" in spec)
apply (drule_tac x="m" in spec)
apply (simp)
apply (erule exE)
apply (simp)

apply (insert normal_limit)
apply (drule_tac x="xps<<i>>" in spec)
apply (drule_tac x="yp i" in spec)
apply (simp add: prod_normal_seq)

apply (drule_tac x="n" in spec)
apply (simp add: symmetry_ms exchange_args_def)

apply (insert ALL_realpow_le_le)
apply (rotate_tac -1)
apply (drule_tac x="1/2" in spec)
apply (drule_tac x="M" in spec)
apply (drule_tac x="n" in spec)
apply (simp_all)
done

(*****************************************
      limit (cms and normal --> limit)
 *****************************************)

lemma prod_cms_normal_limit:
  "ALL (xps::(('i => 'b::cms_rs) infinite_seq)).
      normal xps --> xps convergeTo prod_limit xps"
apply (intro allI impI)
apply (insert prod_limit_to)
apply (drule_tac x="xps" in spec)
apply (drule_tac x="prod_limit xps" in spec)
apply (simp)
apply (drule mp)

apply (intro allI)
apply (simp add: prod_limit_def)
apply (simp add: prod_normal_seq normal_cauchy the_limit)
apply (simp)
done

(*****************************************
      limit (cms and cauchy --> limit)
 *****************************************)

lemma prod_cms_cauchy_limit:
  "ALL (xps::(('i => 'b::cms_rs) infinite_seq)).
      cauchy xps --> xps convergeTo prod_limit (NF xps)"
apply (simp add: prod_cms_normal_limit normal_form_seq_normal 
                 normal_form_seq_same_limit)
done

(*****************************************
     Product Complete Metric Space (RS)
 *****************************************)

lemma prod_cms_rs:
  "ALL (xps::(('i => 'b::cms_rs) infinite_seq)).
       cauchy xps --> (EX yp. xps convergeTo yp)"
apply (intro allI impI)
apply (rule_tac x="prod_limit (NF xps)" in exI)
apply (simp add: prod_cms_cauchy_limit)
done

(************************************************************
                   Product RS ==> CMS and RS
 ************************************************************)

instance fun :: (type,cms_rs) cms
apply (intro_classes)
apply (simp add: prod_cms_rs)
done

instance fun :: (type,cms_rs) cms_rs
apply (intro_classes)
done

(*----------------------------------------------------------*
 |                                                          |
 |                Continuity on Metric space                |
 |                                                          |
 *----------------------------------------------------------*)

lemma prod_continuous:
  "ALL R. (ALL i. continuous (R i)) 
      --> continuous (%xp. (ALL i. (R i) (xp i)))"
apply (intro allI impI)
apply (simp add: continuous_def)
apply (intro allI impI)
apply (erule exE)
apply (drule_tac x="i" in spec)
apply (drule_tac x="x i" in spec)
apply (simp)

apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (drule_tac x="y i" in spec)
apply (rule_tac x="i" in exI)
apply (simp add: prod_restriction_def)
apply (simp add: expand_fun_eq)
done

end

