           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Trace_ren = Prefix:

(*****************************************************************

         1. 
         2. 
         3. 
         4. 

 *****************************************************************)

(*************************************************************
               functions used for defining [[ ]]
 *************************************************************)

consts
  ren_check_list :: "('a list * (('a * 'b) set) * 'b list) => bool"

recdef ren_check_list "measure (%(s,r,t). (length s)+(length t))"
  "ren_check_list ( [], r,  []) = True"
  "ren_check_list (a#s, r, b#t) = ((a,b) : r & ren_check_list (s,r,t))"
  "ren_check_list (  s, r,   t) = False"

consts
  invrelfun :: "('a * 'b) set => 'b set => 'a set" ("(<inv _> _)" [0,1000] 1000)
  addTick   :: "('a * 'b) set => ('a event * 'b event) set"

defs
  invrelfun_def: "<inv r> X == {a. EX b:X. (a,b):r}"
  addTick_def  : "addTick r == {(Ev a, Ev b) |a b. (a,b):r} Un {(Tick, Tick)}"

consts
  ren_check :: "('a trace * (('a * 'b) set) * 'b trace) => bool"

defs
  ren_check_def:
  "ren_check srt == 
   ren_check_list (Rep_trace (fst srt), 
                   addTick (fst (snd srt)), 
                   Rep_trace (snd (snd srt)))"

syntax
  "_invEvrelfun" :: "('a * 'b) set => 'b event set => 'a event set"
                     ("(<invEv _> _)" [0,1000] 1000)

translations
  "<invEv r> X" == "<inv (addTick r)> X"

(*************************************************************
                         lemmas for ren
 *************************************************************)

(******************** list ********************)

(*** decompo <- ***)

lemma ren_decompo_from_list:
  "ALL r t1 t2 s. ren_check_list (s, r, t1 @ t2)
       --> (EX s1 s2. s = s1 @ s2 & ren_check_list (s1, r, t1)
            & ren_check_list (s2, r, t2))"
apply (rule allI)
apply (rule allI)
apply (rule allI)
apply (induct_tac t1)

 (* base *)
apply (rule allI)
apply (intro impI)
apply (rule_tac x="[]" in exI)
apply (rule_tac x="s" in exI)
apply (simp)

 (* step *)
apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (rotate_tac -1)
apply (drule_tac x="s" in spec)
apply (erule disjE)
apply (simp)    (* s=[] --> contradict *)

apply (elim exE)
apply (simp)
apply (drule_tac x="sa" in spec)
apply (simp)
apply (elim exE)

apply (rule_tac x="aa # s1" in exI)
apply (simp)
done

(*** -> decompo ***)

lemma ren_decompo_to_list:
  "ALL r s1 s2 t. ren_check_list (s1 @ s2, r, t)
       --> (EX t1 t2. t = t1 @ t2 & ren_check_list (s1, r, t1)
            & ren_check_list (s2, r, t2))"
apply (rule allI)
apply (rule allI)
apply (rule allI)
apply (induct_tac s1)

 (* base *)
apply (rule allI)
apply (intro impI)
apply (rule_tac x="[]" in exI)
apply (rule_tac x="t" in exI)
apply (simp)

 (* step *)
apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (erule disjE)
apply (simp)    (* t=[] --> contradict *)

apply (elim exE)
apply (simp)
apply (drule_tac x="s" in spec)
apply (simp)
apply (elim exE)

apply (rule_tac x="aa # t1" in exI)
apply (simp)
done

(*** ren append ***)

lemma ren_app_list:
  "ALL r s1 s2 t1 t2.
     ren_check_list (s1, r, t1) & ren_check_list (s2, r, t2)
       --> ren_check_list (s1 @ s2, r, t1 @ t2)"
apply (rule allI)
apply (rule allI)
apply (rule allI)
apply (induct_tac s1)

 (* base *)
apply (intro allI impI)
apply (insert list_nil_or_unnil)
apply (drule_tac x="t1" in spec)
apply (erule disjE)
apply (simp)
apply (elim exE)
apply (simp)    (* s=[] --> contradict *)

 (* step *)
apply (intro allI impI)
apply (rotate_tac 1)
apply (drule_tac x="t1" in spec)
apply (erule disjE)
apply (simp)    (* s=[] --> contradict *)
apply (elim exE)
apply (simp)
done

(*** ren same length ***)

lemma ren_length_list:
  "ALL r s t. ren_check_list (s, r, t) --> length s = length t"
apply (rule allI)
apply (rule allI)
apply (induct_tac s)
apply (rule allI)
apply (simp)
apply (induct_tac t, simp, simp)

 (* step *)
apply (intro allI impI)
apply (insert list_nil_or_unnil)
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (erule disjE)
apply (simp)    (* s=[] --> contradict *)
apply (elim exE)
apply (simp)
done

(******************** trace ********************)

lemma ren_app:
  "ALL r s1 s2 t1 t2.
     s1 @rep s2: trace & t1 @rep t2: trace & 
     ren_check (s1, r, t1) & ren_check (s2, r, t2)
       --> ren_check (s1 @t s2, r, t1 @t t2)"
apply (intro allI impI)
apply (insert ren_app_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="Rep_trace s1" in spec)
apply (drule_tac x="Rep_trace s2" in spec)
apply (drule_tac x="Rep_trace t1" in spec)
apply (drule_tac x="Rep_trace t2" in spec)
apply (simp add: ren_check_def)
apply (simp add: Abs_trace_inverse)
done

lemma ren_app_event:
  "ALL r a b s t.
     (a,b):r & ren_check (s, r, t)
       --> ren_check ([Ev a]t @t s, r, [Ev b]t @t t)"
apply (intro allI impI)
apply (insert ren_app)
apply (drule_tac x="r" in spec)
apply (drule_tac x="[Ev a]t" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Ev b]t" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: trace_event_app_in_trace)
apply (drule mp)
apply (simp add: ren_check_def event_in_trace Abs_trace_inverse)
apply (simp add: addTick_def)
apply (simp)
done

(*** ren same length ***)


lemma ren_length:
  "ALL r s t. ren_check (s, r, t) --> lengtht s = lengtht t"
apply (intro allI impI)
apply (simp add: ren_check_def)
apply (insert ren_length_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (simp)
done

(*** nil nil ***)

lemma ren_nil: "ALL r. ren_check ([]t, r, []t)"
apply (simp add: ren_check_def)
apply (simp add: trace_rep_nil_t)
done

(*** ren Tick ***)

lemma ren_tick: "ALL r. ren_check ([Tick]t, r, [Tick]t)"
apply (simp add: ren_check_def)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: addTick_def)
done

(*** ren [] -> [] ***)

lemma ren_to_nil_lm: 
  "ALL r s. (s : trace & ren_check ([]t, r, Abs_trace s))
      --> (s = [])"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: ren_check_def)
apply (simp add: nil_in_trace Abs_trace_inverse)

apply (insert ren_length_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="[]" in spec)
apply (drule_tac x="s" in spec)
apply (simp)
done

lemma ren_to_nil: 
  "ALL r s. ren_check ([]t, r, s) --> (s = []t)"
apply (intro allI impI)
apply (insert ren_to_nil_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp add: nil_in_trace Abs_Rep_t_inv)
done

(*** ren [] <- [] ***)

lemma ren_from_nil_lm: 
  "ALL r s. (s : trace & ren_check (Abs_trace s, r, []t))
      --> (s = [])"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: ren_check_def)
apply (simp add: nil_in_trace Abs_trace_inverse)

apply (insert ren_length_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[]" in spec)
apply (simp)
done

lemma ren_from_nil: 
  "ALL r s. ren_check (s, r, []t) --> (s = []t)"
apply (intro allI impI)
apply (insert ren_from_nil_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp add: nil_in_trace Abs_Rep_t_inv)
done

(*** ren [Tick] -> [Tick] ***)

lemma ren_to_tick_lm: 
  "ALL r s. (s : trace & ren_check ([Tick]t, r, Abs_trace s))
      --> (s = [Tick])"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: ren_check_def)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert ren_length_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="[Tick]" in spec)
apply (drule_tac x="s" in spec)
apply (simp)

apply (insert list_length_one)
apply (drule_tac x="s" in spec)
apply (simp)
apply (erule exE)
apply (simp add: addTick_def)
done

lemma ren_to_tick: 
  "ALL r s. ren_check ([Tick]t, r, s) --> (s = [Tick]t)"
apply (intro allI impI)
apply (insert ren_to_tick_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp add: event_in_trace Abs_Rep_t_inv)
done

(*** ren [Tick] <- [Tick] ***)

lemma ren_from_tick_lm: 
  "ALL r s. (s : trace & ren_check (Abs_trace s, r, [Tick]t))
      --> (s = [Tick])"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: ren_check_def)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert ren_length_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Tick]" in spec)
apply (simp)

apply (insert list_length_one)
apply (drule_tac x="s" in spec)
apply (simp)
apply (erule exE)
apply (simp add: addTick_def)
done

lemma ren_from_tick: 
  "ALL r s. ren_check (s, r, [Tick]t) --> (s = [Tick]t)"
apply (intro allI impI)
apply (insert ren_from_tick_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp add: event_in_trace Abs_Rep_t_inv)
done

(*** ren [Ev a] <- [Ev b] ***)

lemma ren_from_Ev_lm: 
  "ALL r s b. (s : trace & ren_check (Abs_trace s, r, [Ev b]t))
      --> (EX a. s = [Ev a] & (a,b):r)"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: ren_check_def)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert ren_length_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Ev b]" in spec)
apply (simp)

apply (insert list_length_one)
apply (drule_tac x="s" in spec)
apply (simp)
apply (erule exE)
apply (simp add: addTick_def)
done

lemma ren_from_Ev: 
  "ALL r s b. ren_check (s, r, [Ev b]t)
      --> (EX a. s = [Ev a]t & (a,b):r)"
apply (intro allI impI)
apply (insert ren_from_Ev_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="b" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp add: event_in_trace Abs_Rep_t_inv)
done

(*** ren [Ev a] -> [Ev b] ***)

lemma ren_to_Ev_lm: 
  "ALL r s a. (s : trace & ren_check ([Ev a]t, r, Abs_trace s))
      --> (EX b. s = [Ev b] & (a,b):r)"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: ren_check_def)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert ren_length_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="[Ev a]" in spec)
apply (drule_tac x="s" in spec)
apply (simp)

apply (insert list_length_one)
apply (drule_tac x="s" in spec)
apply (simp)
apply (erule exE)
apply (simp add: addTick_def)
done

lemma ren_to_Ev: 
  "ALL r s a. ren_check ([Ev a]t, r, s)
      --> (EX b. s = [Ev b]t & (a,b):r)"
apply (intro allI impI)
apply (insert ren_to_Ev_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="a" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp add: event_in_trace Abs_Rep_t_inv)
done

(*** ren notick <- notick ***)

lemma ren_from_notick_lm: 
  "ALL r t s. (s : trace & t : trace & Tick ~: set t &
               ren_check (Abs_trace s, r, Abs_trace t))
           --> Tick ~: set s"
apply (rule allI)
apply (rule allI)

apply (induct_tac t)
apply (intro allI impI)
apply (elim conjE)
apply (insert ren_from_nil_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="s" in spec)
apply (simp)

apply (intro allI impI)
apply (insert list_nil_or_unnil)
apply (rotate_tac -1)
apply (drule_tac x="s" in spec)
apply (erule disjE)
apply (simp)

apply (elim exE conjE)
apply (simp add: ren_check_def Abs_trace_inverse)

apply (drule_tac x="sa" in spec)
apply (drule mp)
apply (insert event_app_in_trace_inv)
apply (drule_tac x="a" in spec)
apply (drule_tac x="list" in spec)
apply (insert event_app_in_trace_inv)
apply (drule_tac x="aa" in spec)
apply (drule_tac x="sa" in spec)
apply (simp add: Abs_trace_inverse)

apply (simp)
apply (case_tac "aa = Tick")
apply (simp add: addTick_def)
apply (auto)
done

(*** notick <-- ***)

lemma ren_from_notick: 
  "ALL r t s. notick t & ren_check (s, r, t) --> notick s"
apply (intro allI impI)
apply (insert ren_from_notick_lm)
apply (drule_tac x="r" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
done

(*** ren Prefix ***)

lemma ren_prefix:
  "ALL r v u s. prefix v u & ren_check (s, r, u)
     --> (EX t. prefix t s & ren_check (t, r, v))"
apply (intro allI impI)
apply (simp add: prefix_list_def)
apply (elim conjE exE)

apply (simp add: ren_check_def Rep_trace)
apply (insert ren_decompo_from_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="Rep_trace v" in spec)
apply (drule_tac x="ua" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp)

apply (elim conjE exE)
apply (rule_tac x="Abs_trace s1" in exI)

apply (rotate_tac 3)
apply (drule sym)

apply (insert app_trace_in_trace_inv)
apply (drule_tac x="s1" in spec)
apply (drule_tac x="s2" in spec)
apply (simp add: Rep_trace Abs_trace_inverse)
apply (rule_tac x="s2" in exI)
apply (simp)
done

(*** ren decompo <- t1 @t t2 ***)

lemma ren_decompo_from:
  "ALL r t1 t2 s. notick t1 & ren_check (s, r, t1 @t t2)
       --> (EX s1 s2. s = s1 @t s2 & ren_check (s1, r, t1)
                                   & ren_check (s2, r, t2))"
apply (intro allI impI)
apply (simp add: ren_check_def)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="t1" in spec)
apply (drule_tac x="t2" in spec)
apply (simp add: Abs_trace_inverse)

apply (insert ren_decompo_from_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="Rep_trace t1" in spec)
apply (drule_tac x="Rep_trace t2" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp)
apply (elim conjE exE)

apply (rule_tac x="Abs_trace s1" in exI)
apply (rule_tac x="Abs_trace s2" in exI)

apply (rotate_tac 3)
apply (drule sym)

apply (insert app_trace_in_trace_inv2)
apply (drule_tac x="s1" in spec)
apply (drule_tac x="s2" in spec)
apply (simp add: Rep_trace)

apply (simp add: Abs_trace_inverse Rep_trace_inverse)
done

(*** ren s1 @t s2 --> decompo ***)

lemma ren_decompo_to:
  "ALL r s1 s2 t. notick s1 & ren_check (s1 @t s2, r, t)
       --> (EX t1 t2. t = t1 @t t2 & ren_check (s1, r, t1)
                                   & ren_check (s2, r, t2))"
apply (intro allI impI)
apply (simp add: ren_check_def)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="s1" in spec)
apply (drule_tac x="s2" in spec)
apply (simp add: Abs_trace_inverse)

apply (insert ren_decompo_to_list)
apply (drule_tac x="addTick r" in spec)
apply (drule_tac x="Rep_trace s1" in spec)
apply (drule_tac x="Rep_trace s2" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (simp)
apply (elim conjE exE)

apply (rule_tac x="Abs_trace t1" in exI)
apply (rule_tac x="Abs_trace t2" in exI)

apply (rotate_tac 3)
apply (drule sym)

apply (insert app_trace_in_trace_inv2)
apply (drule_tac x="t1" in spec)
apply (drule_tac x="t2" in spec)
apply (simp add: Rep_trace)

apply (simp add: Abs_trace_inverse Rep_trace_inverse)
done

(*************************************************************
                       inverse R
 *************************************************************)

lemma invEv_sub_Evset: "ALL r. <invEv r> Evset <= Evset"
apply (auto simp add: invrelfun_def addTick_def Evset_def)
done

lemma invrelfun_sub:
  "ALL r X Y. X <= Y --> <inv r> X <= <inv r> Y"
apply (auto simp add: invrelfun_def)
done

lemma invrelfun_Un:
  "ALL r X Y. <inv r> X Un <inv r> Y = <inv r> (X Un Y)"
apply (auto simp add: invrelfun_def)
done

(*** <invEv r> X <= Evset ***)

lemma invEv_sub_X_Evset_only_if: "ALL r X. <invEv r> X <= Evset --> X <= Evset"
apply (intro allI impI)
apply (case_tac "Tick : X")  (* --> contra *)
apply (simp add: addTick_def invrelfun_def)
apply (erule subsetE)
apply (drule_tac x="Tick" in bspec, simp)
apply (simp add: Evset_def)

apply (simp add: Evset_def)
apply (force)
done

lemma invEv_sub_X_Evset_if: "ALL r X. X <= Evset --> <invEv r> X <= Evset"
apply (intro allI impI)
apply (simp add: Evset_def)
apply (simp add: addTick_def invrelfun_def)
apply (auto)
done

lemma invEv_sub_X_Evset: "ALL r X. (<invEv r> X <= Evset) = (X <= Evset)"
apply (intro allI)
apply (rule iffI)
apply (insert invEv_sub_X_Evset_only_if)
apply (drule_tac x="r" in spec)
apply (drule_tac x="X" in spec)
apply (simp)
apply (simp add: invEv_sub_X_Evset_if)
done

end


