           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Trace_seq = Prefix:

(*****************************************************************

         1. 
         2. 
         3. 
         4. 

 *****************************************************************)

consts
  rmtick_list :: "'a event list => 'a event list"

primrec
  "rmtick_list []    = []"
  "rmtick_list (a#s) = (if a=Tick then rmtick_list s 
                                  else a#(rmtick_list s))"

syntax
  "_rmtick"  :: "'a trace => 'a trace"              ("rmtick")

translations
  "rmtick s"  == "Abs_trace (rmtick_list (Rep_trace s))"

(***********************************************************
                          lemmas
 ***********************************************************)

(********************** list **********************)

lemma rmtick_notick_list:
  "ALL s. Tick ~: set s --> rmtick_list s = s"
apply (rule allI)
apply (induct_tac s)
apply (simp)
apply (intro impI)
apply (auto)
done

(*** rmtick distribution ***)

lemma rmtick_app_dist_list:
  "ALL s t. rmtick_list (s @ t) = (rmtick_list s) @ (rmtick_list t)"
apply (rule allI)
apply (induct_tac s)
apply (auto)
done

(*** rmtick s = s | (rmtick s) @t [Tick]t = s (list) ***)

lemma rmtick_trace_list: 
  "ALL s. s:trace --> rmtick_list s = s |  (rmtick_list s) @ [Tick] = s"
apply (rule allI)
apply (simp add: trace_def)
apply (insert list_last_nil_or_unnil)
apply (drule_tac x="s" in spec)
apply (erule disjE)
apply (simp)

apply (elim exE)
apply (simp)
apply (intro impI)

apply (case_tac "a=Tick")
apply (rule disjI2)
apply (simp add: rmtick_app_dist_list)
apply (simp add: rmtick_notick_list)

apply (rule disjI1)
apply (simp add: rmtick_app_dist_list)
apply (simp add: rmtick_notick_list)
done

(*** rmtick prefix ***)

lemma rmtick_prefix_list: 
  "ALL s. s:trace --> prefix_list (rmtick_list s) s"
apply (intro allI impI)
apply (simp add: prefix_list_def)
apply (insert rmtick_trace_list)
apply (drule_tac x="s" in spec)
apply (simp)
apply (erule disjE)
apply (rule_tac x="[]" in exI)
apply (simp)
apply (rule_tac x="[Tick]" in exI)
apply (simp)
done

(*** notick rmtick_list ***)

lemma rmtick_is_notick_list:
  "ALL s. Tick ~: set (rmtick_list s)"
apply (rule allI)
apply (induct_tac s)
apply (auto)
done

(*** rmtick_list t : trace ***)

lemma rmtick_list_in_trace: "ALL t. rmtick_list t : trace"
apply (intro allI)
apply (simp add: trace_def)
apply (insert butlast_subseteq)
apply (drule_tac x="rmtick_list t" in spec)
apply (auto simp add: rmtick_is_notick_list)
done

(*** rmtick s = t1 @ t2 --> s = s1 @ s2 ***)

lemma rmtick_from_decompo_list:
  "ALL s t1 t2. rmtick_list s = t1 @ t2
        --> (EX s1 s2. s = s1 @ s2 &
                       rmtick_list s1 = t1 & rmtick_list s2 = t2)"
apply (rule allI)
apply (induct_tac s)

 (* base *)
apply (intro allI impI)
apply (rule_tac x="[]" in exI)
apply (rule_tac x="[]" in exI)
apply (simp)

 (* step "s = a # list" *)
apply (intro allI impI)

 apply (case_tac "a = Tick")
 apply (drule_tac x="t1" in spec)
 apply (drule_tac x="t2" in spec)
 apply (simp)
 apply (elim conjE exE)
 apply (rule_tac x="a#s1" in exI)
 apply (rule_tac x="s2" in exI)
 apply (simp)

 (* a ~= Tick *)

 apply (simp)

 apply (insert list_nil_or_unnil)
 apply (rotate_tac -1)
 apply (drule_tac x="t1" in spec)
 apply (erule disjE)

  (* t1 = [] *)
 apply (simp)

 apply (insert list_nil_or_unnil)
 apply (rotate_tac -1)
 apply (drule_tac x="t2" in spec)
 apply (erule disjE)

   (* t2 = [] *)
  apply (simp)        (* --> contra *)

   (* t2 = aa # s *)
  apply (elim exE)

  apply (drule_tac x="[]" in spec)
  apply (drule_tac x="s" in spec)
  apply (simp)
  apply (elim exE)

  apply (rule_tac x="[]" in exI)
  apply (rule_tac x="a # s1 @ s2" in exI)
  apply (simp)

  (* t1 ~= [] *)

 apply (elim exE)
 apply (drule_tac x="s" in spec)
 apply (drule_tac x="t2" in spec)
 apply (simp)
 apply (elim conjE exE)

 apply (rule_tac x="a # s1" in exI)
 apply (rule_tac x="s2" in exI)
 apply (simp)
done

(********************** trace **********************)

(*** notick --> rmtick s ***)

lemma rmtick_notick: "ALL s. notick s --> rmtick s = s"
apply (simp add: rmtick_notick_list Rep_trace_inverse)
done

(*** rmtick []t ***)

lemma rmtick_nil: "rmtick []t = []t"
apply (simp add: rmtick_notick notick_nil)
done

(*** rmtick [Tick]t ***)

lemma rmtick_tick: "rmtick [Tick]t = []t"
apply (simp add: event_in_trace Abs_trace_inverse)
done

(*** notick rmtick s ***)

lemma rmtick_is_notick: "ALL s. notick (rmtick s)"
apply (rule allI)
apply (insert rmtick_is_notick_list)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: rmtick_list_in_trace Abs_trace_inverse)
done

(*** rmtick distribution ***)

lemma rmtick_app_dist:
  "ALL s t. notick s --> rmtick (s @t t) = (rmtick s) @t (rmtick t)"
apply (intro allI impI)
apply (insert rmtick_app_dist_list)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (simp add: trace_app_in_trace_if Abs_trace_inverse)
apply (simp add: rmtick_is_notick_list rmtick_is_notick trace_app_in_trace_if
                 app_trace_notick rmtick_list_in_trace 
                 Abs_trace_inverse Abs_trace_inject)
done

(*** rmtick distribution on event ***)

lemma rmtick_app_dist_event:
  "ALL a t. rmtick ([Ev a]t @t t) = [Ev a]t @t (rmtick t)"
apply (intro allI impI)
apply (insert rmtick_app_dist)
apply (drule_tac x="[Ev a]t" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: event_in_trace Abs_trace_inverse)
done

(*** rmtick s = s | (rmtick s) @t [Tick]t = s ***)

lemma rmtick_trace: 
  "ALL s. rmtick s = s | (rmtick s) @t [Tick]t = s"
apply (rule allI)
apply (insert rmtick_trace_list)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: Rep_trace)
apply (erule disjE)

apply (drule sym)
apply (insert Abs_Rep_t_inv_imp)
apply (drule_tac x="s" in spec)
apply (drule_tac x="rmtick_list (Rep_trace s)" in spec)
apply (simp)

apply (rule disjI2)
apply (simp add: event_in_trace rmtick_list_in_trace Abs_trace_inverse)
apply (simp add: Rep_trace_inverse)
done

(*** rmtick prefix ***)

lemma rmtick_prefix:
  "ALL s t. prefix s (rmtick t) --> (EX u. s = rmtick u & prefix u t)"
apply (intro allI impI)
apply (simp add: prefix_list_def)
apply (elim conjE exE)
apply (simp add: rmtick_list_in_trace Abs_trace_inverse)

apply (insert rmtick_from_decompo_list)
apply (drule_tac x="(Rep_trace t)" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="u" in spec)
apply (simp)
apply (elim conjE exE)

apply (rule_tac x="Abs_trace s1" in exI)

apply (rotate_tac -2)
apply (drule sym)

apply (insert Abs_Rep_t_inv_imp)
apply (drule_tac x="s" in spec)
apply (drule_tac x="rmtick_list s1" in spec)

apply (simp)
apply (insert app_trace_in_trace_inv)
apply (drule_tac x="s1" in spec)
apply (drule_tac x="s2" in spec)

apply (rotate_tac 3)
apply (drule sym)
apply (simp add: Rep_trace Abs_trace_inverse)
apply (rule_tac x="s2" in exI)
apply (simp)
done

(*** rmtick prefix ***)

lemma rmtick_is_prefix: "ALL s. prefix (rmtick s) s"
apply (intro allI)

apply (insert rmtick_trace)
apply (drule_tac x="s" in spec)
apply (erule disjE)
apply (simp add: prefix_list_def)
apply (simp add: Rep_trace)

apply (simp add: prefix_list_def)
apply (simp add: Rep_trace)
apply (rule_tac x="Rep_trace [Tick]t" in exI)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="rmtick s" in spec)
apply (drule_tac x="[Tick]t" in spec)
apply (simp add: rmtick_is_notick)
apply (simp add: Abs_Rep_t)
done

end
