           (*-------------------------------------------*
            |  The Dining Mathematicians in Csp-Prover  |
            |               August 2004                 |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory DM4_Spc_def = DM3_hide:

(*****************************************************************

         1. defines Spc 
         2.
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                     specification 
 *********************************************************)

datatype SpcName = TH0_TH1 | EAT0_TH1 | TH0_EAT1

consts
  SpcDef :: "(SpcName, Event) procDef"

primrec
   "SpcDef
      TH0_TH1  = ! x:OBS -> IF (x = Eat0) THEN <EAT0_TH1>
                                          ELSE IF (x = Eat1) 
                                               THEN <TH0_EAT1>
                                               ELSE <TH0_TH1>"

   "SpcDef
      EAT0_TH1 = ! x:(OBS - {Eat1}) -> IF (x = End0) 
                                       THEN <TH0_TH1>
                                       ELSE <EAT0_TH1>"

   "SpcDef
      TH0_EAT1 = ! x:(OBS - {Eat0}) -> IF (x = End1) 
                                       THEN <TH0_TH1>
                                       ELSE <TH0_EAT1>"

consts
  Spc :: "(SpcName, Event) procRec"

defs
  Spc_def:
   "Spc == LET SpcDef IN <TH0_TH1>"


(*********************************************************
        Lemmas for   ALL n.  verify SPC <=sf SYS n
 *********************************************************)

(*** relation between SpcDef and SyDef ***)

consts 
  Spc_to_Sys :: "SpcName => (SysName, Event) proc"

primrec
  "Spc_to_Sys
     TH0_TH1 = 
       NUM!!n .. (<TH0> |[CH0]| <VAR n> |[CH1]| <TH1>) -- (CH0 Un CH1)
       |~|
       NUM!!n .. (Back0 -> <TH0> |[CH0]| <VAR n> |[CH1]| <TH1>) -- (CH0 Un CH1)
       |~|
       NUM!!n .. (<TH0> |[CH0]| <VAR n> |[CH1]| Back1 -> <TH1>) -- (CH0 Un CH1)
       |~|
       NUM!!n .. (Back0 -> <TH0> |[CH0]| <VAR n> |[CH1]| Back1 -> <TH1>) 
                  -- (CH0 Un CH1)
       |~|
       NUM!!n:EVENs .. (Eat0 -> <EAT0 n> |[CH0]| <VAR n> |[CH1]| Back1 -> <TH1>)
                        -- (CH0 Un CH1)
       |~|
       NUM!!n:ODDs  .. (Back0 -> <TH0> |[CH0]| <VAR n> |[CH1]| Eat1 -> <EAT1 n>)
                        -- (CH0 Un CH1)
       |~|
       NUM!!n:EVENs .. (Eat0 -> <EAT0 n> |[CH0]| <VAR n> |[CH1]| <TH1>)
                        -- (CH0 Un CH1)
       |~|
       NUM!!n:ODDs  .. (<TH0> |[CH0]| <VAR n> |[CH1]| Eat1 -> <EAT1 n>)
                        -- (CH0 Un CH1)
       |~|
       NUM!!n:EVENs .. (WR0 (n div 2) -> <TH0> |[CH0]| <VAR n> |[CH1]| <TH1>)
                       -- (CH0 Un CH1)
       |~|
       NUM!!n:EVENs .. (WR0 (n div 2) -> <TH0> |[CH0]| <VAR n> |[CH1]| Back1 -> <TH1>)
                       -- (CH0 Un CH1)
       |~|
       NUM!!n:ODDs  .. (<TH0> |[CH0]| <VAR n> |[CH1]| WR1 ! (3 * n + 1) -> <TH1>)
                       -- (CH0 Un CH1)
       |~|
       NUM!!n:ODDs  .. (Back0 -> <TH0> |[CH0]| <VAR n> |[CH1]| WR1 ! (3 * n + 1) -> <TH1>)
                       -- (CH0 Un CH1)"

  "Spc_to_Sys
     EAT0_TH1 = 
       NUM!!n:EVENs .. (<EAT0 n> |[CH0]| <VAR n> |[CH1]| <TH1>) -- (CH0 Un CH1)
       |~|
       NUM!!n:EVENs .. (<EAT0 n> |[CH0]| <VAR n> |[CH1]| Back1 -> <TH1>)
                       -- (CH0 Un CH1)"

  "Spc_to_Sys
     TH0_EAT1 = 
       NUM!!n:ODDs .. (<TH0> |[CH0]| <VAR n> |[CH1]| <EAT1 n>) -- (CH0 Un CH1)
       |~|
       NUM!!n:ODDs .. (Back0 -> <TH0> |[CH0]| <VAR n> |[CH1]| <EAT1 n>)
                       -- (CH0 Un CH1)"

(*********************************************************
               nohide and guarded lemmas
 *********************************************************)

lemma guardSpcDef[simp]: "!!C. guard ((SpcDef) C)"
by (induct_tac C, simp_all)

lemma nohideSpcDef[simp]: "!!C. nohide ((SpcDef) C)"
by (induct_tac C, simp_all)

end
