           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                August 2004                |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Inc_nat = Csp_Prover:

(*****************************************************************

         1. simple example for fixed point inductuction theorem
         2. Parallel, Hiding, Internal choice
         3. Refinement
         4. 

 *****************************************************************)

(*********************************************************
                         event
 *********************************************************)

datatype Event = Num nat | Read nat
datatype SpcName = Cspc nat
datatype SysName = UI | VAR nat

(*********************************************************
            specification SPC and system SYS
 *********************************************************)

consts
  GTs :: "nat => nat set"

defs
  GTs_def : "GTs n == {m. n < m}"

consts 
  SpcDef :: "(SpcName, Event) procDef"
  SysDef :: "(SysName, Event) procDef"

primrec
  "SpcDef  (Cspc n) = Num n -> Num !! x:(GTs n) .. <Cspc x>"

primrec
  "SysDef      (UI) = Read ? m -> Num m -> <UI>"
  "SysDef   (VAR n) = Read ! n -> <VAR (Suc n)>"

consts 
  Spc :: "(SpcName, Event) procRec"
  Sys :: "(SysName, Event) procRec"

defs
  Spc_def: "Spc == LET SpcDef IN <Cspc 0>"

defs
  Sys_def: "Sys == LET SysDef IN (<UI> |[range Read]| <VAR 0>) -- (range Read)"

(*********************************************************
            relation between SPC and SYS
 *********************************************************)

consts 
  Spc_to_Sys :: "SpcName => (SysName, Event) proc"

primrec
  "Spc_to_Sys (Cspc n) = (<UI> |[range Read]| <VAR n>) -- (range Read)"

(*********************************************************
                     small lemmas
 *********************************************************)

declare Int_commute [simp]

lemma set1[simp]: "range Read Int {Read n} = {Read n}"
apply (auto)
done

lemma set2[simp]: "{Read (Suc n)} Int (range Read Int {Num n}) Un
                    ({Num n} - range Read) = {Num n}"
apply (auto)
done

lemma set3[simp]: "Num n ~: range Read"
apply (simp add: image_def)
done

(*********************************************************
               nohide and guarded lemmas
 *********************************************************)

(*** To automatically unfold syntactic sugar ***)
declare csp_ss_def[simp]

lemma guardSysDef[simp]: "!!C. guard ((SysDef) C)"
apply (induct_tac C, simp_all)
done

lemma nohideSysDef[simp]: "!!C. nohide ((SysDef) C)"
apply (induct_tac C, simp_all)
done

lemma guardSpcDef[simp]: "!!C. guard ((SpcDef) C)"
apply (induct_tac C, simp_all)
done

lemma nohideSpcDef[simp]: "!!C. nohide ((SpcDef) C)"
apply (induct_tac C, simp_all)
done

(*********************************************************
                   ? SPC <=sf SYS ?
 *********************************************************)

lemma "Spc <=F Sys"
  (***** by fixed point induction *****)
apply (rule csp_fp_induct[of _ _ "Spc_to_Sys"])
apply (rule)

  (***** check guarded and no hiding operators *****)
apply (simp_all add: Spc_def Sys_def)
apply (induct_tac C)
apply (simp)

  (***** step (Parallel and Hidden) *****)

apply (tactic {* csp_hnf_tac 1 *})
apply (auto)
apply (tactic {* csp_simp_tac 1 *})

  (***** final step *****)

apply (rule csp_Int_bind_left)
apply (rule_tac  x="Num (Suc nat)" in exI)
apply (simp add: GTs_def)

apply (tactic {* csp_unwind_tac 1 *})
apply (tactic {* csp_light_step_tac 1 *})
done

end
